﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Configuration {
  using System.Collections.Generic;

  using HeuristicLab.Common;
  using HeuristicLab.Core;
  using HeuristicLab.Data;
  using HeuristicLab.Parameters;
  using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
  using HeuristicLab.Problems.ProgramSynthesis.Base.Erc;

  [StorableClass]
  public class PushConfigurationParameterCollection : PushConfigurationBase, IReadOnlyPushConfiguration {
    private const string InstructionsParameterName = "Instructions";
    private const string InstructionsParameterDescription = "Enables/Disables Instructions";
    private const string EvalPushLimitParameterName = "EvalPushLimit";
    private const string EvalPushLimitParameterDescription = "This is the maximum allowed number of \"executions\" in a single top-level call to the interpreter. The execution of a single Push instruction counts as one execution, as does the processing of a single literal, as does the descent into one layer of parentheses (that is, the processing of the \"(\" counts as one execution).";
    private const string MaxPointsInProgramParameterName = "MaxProgramLength";
    private const string MaxProgramLengthParameterDescription = "This is the maximum size of an item on the CODE/EXEC stack, expressed as a number of points. A point is an instruction, a literal, or a pair of parentheses.";
    private const string TopLevelPushCodeParameterName = "TopLevelPushCode";
    private const string TopLevelPushCodeParameterDescription = "When TRUE (which is the default), code passed to the top level of the interpreter will be pushed onto the CODE stack prior to execution.";
    private const string TopLevelPopCodeParameterName = "TopLevelPopCode";
    private const string TopLevelPopCodeParameterDescription = "When TRUE, the CODE stack will be popped at the end of top level calls to the interpreter. The default is FALSE.";
    private const string MaxPointsInRandomInstructionParameterName = "MaxPointsInRandomInstruction";
    private const string MaxPointsInRandomInstructionParameterDescription = "MaxPointsInRandomInstruction";
    private const string ErcOptionsParameterName = "ERC options";
    private const string MaxStringLengthParameterName = "Max. string length";
    private const string MaxDepthParameterName = "Max. program recursion";
    private const string MaxParenthesesCloseParameterName = "Max. parentheses close";
    private const string MaxParenthesesCloseParameterDescription = "Specifies how many sub programs are closed if open during the recursive translation of an individual to a push program. Value is exclusive.";
    private const string ParenthesesCloseBiasLevelParameterName = "Parentheses close bias level";
    private const string ParenthesesCloseBiasLevelParameterDescription = "Specifies how strong a random value between 0 .. 'Max. parentheses close' is biased towards 0. In other words, this parameter controls the length of sub programs.";
    private const string MaxVectorLengthParameterName = "Max. vector length";

    public PushConfigurationParameterCollection() {
      Parameters = new ParameterCollection();

      InitParameters();
    }

    public PushConfigurationParameterCollection(PushConfigurationParameterCollection origin, Cloner cloner) : base(origin, cloner) {
      Parameters = cloner.Clone(origin.Parameters);
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new PushConfigurationParameterCollection(this, cloner);
    }

    [StorableConstructor]
    protected PushConfigurationParameterCollection(bool deserialize) { }

    [StorableHook(HookType.AfterDeserialization)]
    // ReSharper disable once UnusedMember.Local
    private void AfterDeserialization() {
      InitParameters();
    }

    [Storable]
    public ParameterCollection Parameters { get; private set; }

    [Storable]
    public string FloatStringFormat { get; set; }

    private void InitParameters() {
      if (!Parameters.ContainsKey(InstructionsParameterName))
        Parameters.Add(new ValueParameter<IEnabledExpressionsConfiguration>(
          InstructionsParameterName,
          InstructionsParameterDescription,
          this));

      if (!Parameters.ContainsKey(ErcOptionsParameterName))
        Parameters.Add(new ValueParameter<ErcOptions>(ErcOptionsParameterName));

      if (!Parameters.ContainsKey(MaxVectorLengthParameterName))
        Parameters.Add(new FixedValueParameter<IntValue>(
          MaxVectorLengthParameterName,
          new IntValue(500)) { Hidden = true });

      if (!Parameters.ContainsKey(EvalPushLimitParameterName))
        Parameters.Add(new FixedValueParameter<IntValue>(
          EvalPushLimitParameterName,
          EvalPushLimitParameterDescription,
          new IntValue(1000)));

      if (!Parameters.ContainsKey(MaxPointsInProgramParameterName))
        Parameters.Add(new FixedValueParameter<IntValue>(
          MaxPointsInProgramParameterName,
          MaxProgramLengthParameterDescription,
          new IntValue(200)));

      if (!Parameters.ContainsKey(MaxParenthesesCloseParameterName))
        Parameters.Add(new FixedValueParameter<IntValue>(
          MaxParenthesesCloseParameterName,
          MaxParenthesesCloseParameterDescription,
          new IntValue(4)) { Hidden = true });

      if (!Parameters.ContainsKey(ParenthesesCloseBiasLevelParameterName))
        Parameters.Add(new FixedValueParameter<DoubleValue>(
          ParenthesesCloseBiasLevelParameterName,
          ParenthesesCloseBiasLevelParameterDescription,
          new DoubleValue(4)) { Hidden = true });

      if (!Parameters.ContainsKey(TopLevelPushCodeParameterName))
        Parameters.Add(new FixedValueParameter<BoolValue>(
          TopLevelPushCodeParameterName,
          TopLevelPushCodeParameterDescription,
          new BoolValue(true)) { Hidden = true });

      if (!Parameters.ContainsKey(TopLevelPopCodeParameterName))
        Parameters.Add(new FixedValueParameter<BoolValue>(
          TopLevelPopCodeParameterName,
          TopLevelPopCodeParameterDescription,
          new BoolValue(false)) { Hidden = true });

      if (!Parameters.ContainsKey(MaxPointsInRandomInstructionParameterName))
        Parameters.Add(new FixedValueParameter<IntValue>(
          MaxPointsInRandomInstructionParameterName,
          MaxPointsInRandomInstructionParameterDescription,
          new IntValue(50)) { Hidden = true });

      if (!Parameters.ContainsKey(MaxStringLengthParameterName))
        Parameters.Add(new FixedValueParameter<IntValue>(
          MaxStringLengthParameterName,
          new IntValue(1000)) { Hidden = true });

      if (!Parameters.ContainsKey(MaxDepthParameterName))
        Parameters.Add(new FixedValueParameter<IntValue>(
          MaxDepthParameterName,
          new IntValue(1000)) { Hidden = true });
    }

    public IValueParameter<IEnabledExpressionsConfiguration> InstructionsParameter
    {
      get { return (IValueParameter<IEnabledExpressionsConfiguration>)Parameters[InstructionsParameterName]; }
    }

    public IEnabledExpressionsConfiguration Instructions
    {
      get { return InstructionsParameter.Value; }
      set { InstructionsParameter.Value = value; }
    }

    public IValueParameter<ErcOptions> ErcOptionsParameter
    {
      get { return (IValueParameter<ErcOptions>)Parameters[ErcOptionsParameterName]; }
    }

    public ErcOptions ErcOptions
    {
      get { return ErcOptionsParameter.Value; }
      set
      {
        ErcOptionsParameter.Value = value;
      }
    }

    IReadOnlyList<string> IReadOnlyPushConfiguration.EnabledExpressions
    {
      get
      {
        return enabledExpressions;
      }
    }

    IReadOnlyErcOptions IReadOnlyPushConfiguration.ErcOptions
    {
      get
      {
        return ErcOptions;
      }
    }

    /// <summary>
    ///     This is the maximum allowed number of "executions" in a single top-level call to the interpreter.
    ///     The execution of a single Push instruction counts as one execution, as does the processing of a single literal,
    ///     as does the descent into one layer of parentheses (that is, the processing of the "(" counts as one execution).
    ///     When this limit is exceeded the interpreter aborts immediately, leaving its stacks in the states they were in prior
    ///     to the abort (so they may still be examined by a calling program). Whether or not this counts as an "abnormal"
    ///     termination
    ///     is up to the calling program.
    /// </summary>
    public IValueParameter<IntValue> EvalPushLimitParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[EvalPushLimitParameterName]; }
    }

    public int EvalPushLimit
    {
      get { return EvalPushLimitParameter.Value.Value; }
      set { EvalPushLimitParameter.Value.Value = value; }
    }


    /// <summary>
    /// This is the maximum of depth a push program can have. Expressions, which lead to exceed this limit are interpreted as NOOP.
    /// </summary>
    public IValueParameter<DoubleValue> ParenthesesCloseBiasLevelParameter
    {
      get { return (IValueParameter<DoubleValue>)Parameters[ParenthesesCloseBiasLevelParameterName]; }
    }

    public double ParenthesesCloseBiasLevel
    {
      get { return ParenthesesCloseBiasLevelParameter.Value.Value; }
      set { ParenthesesCloseBiasLevelParameter.Value.Value = value; }
    }

    /// <summary>
    /// This is the maximum of depth a push program can have. Expressions, which lead to exceed this limit are interpreted as NOOP.
    /// </summary>
    public IValueParameter<IntValue> MaxParenthesesCloseParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[MaxParenthesesCloseParameterName]; }
    }

    public int MaxParenthesesClose
    {
      get { return MaxParenthesesCloseParameter.Value.Value; }
      set
      {
        MaxParenthesesCloseParameter.Value.Value = value;
      }
    }

    /// <summary>
    /// This is the maximum of depth a push program can have. Expressions, which lead to exceed this limit are interpreted as NOOP.
    /// </summary>
    public IValueParameter<IntValue> MaxDepthParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[MaxDepthParameterName]; }
    }

    public int MaxDepth
    {
      get { return MaxDepthParameter.Value.Value; }
      set
      {
        MaxDepthParameter.Value.Value = value;
      }
    }

    /// <summary>
    ///     This is the maximum size of an item on the CODE stack, expressed as a number of points.
    ///     A point is an instruction, a literal, or a pair of parentheses. Any instruction that would cause this limit to be
    ///     exceeded should instead act as a NOOP, leaving all stacks in the states that they were in before the execution of the
    ///     instruction.
    /// </summary>
    public IValueParameter<IntValue> MaxPointsInProgramParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[MaxPointsInProgramParameterName]; }
    }

    public int MaxPointsInProgram
    {
      get { return MaxPointsInProgramParameter.Value.Value; }
      set
      {
        MaxPointsInProgramParameter.Value.Value = value;
      }
    }

    public IValueParameter<IntValue> MaxVectorLengthParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[MaxVectorLengthParameterName]; }
    }

    public int MaxVectorLength
    {
      get { return MaxVectorLengthParameter.Value.Value; }
      set { MaxVectorLengthParameter.Value.Value = value; }
    }

    /// <summary>
    ///     The maximum number of points in an expression produced by the CODE.RAND instruction.
    /// </summary>
    public IValueParameter<IntValue> MaxPointsInRandomExpressionParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[MaxPointsInRandomInstructionParameterName]; }
    }

    public int MaxPointsInRandomExpression
    {
      get { return MaxPointsInRandomExpressionParameter.Value.Value; }
      set
      {
        MaxPointsInRandomExpressionParameter.Value.Value = value;
      }
    }

    /// <summary>
    ///     When TRUE (which is the default), code passed to the top level of the interpreter
    ///     will be pushed onto the CODE stack prior to execution.
    /// </summary>
    public IValueParameter<BoolValue> TopLevelPushCodeParameter
    {
      get { return (IValueParameter<BoolValue>)Parameters[TopLevelPushCodeParameterName]; }
    }

    public bool TopLevelPushCode
    {
      get { return TopLevelPushCodeParameter.Value.Value; }
      set
      {
        TopLevelPushCodeParameter.Value.Value = value;
      }
    }

    /// <summary>
    ///     When TRUE, the CODE stack will be popped at the end of top level calls to the interpreter. The default is FALSE.
    /// </summary>
    public IValueParameter<BoolValue> TopLevelPopCodeParameter
    {
      get { return (IValueParameter<BoolValue>)Parameters[TopLevelPopCodeParameterName]; }
    }

    public bool TopLevelPopCode
    {
      get { return TopLevelPopCodeParameter.Value.Value; }
      set
      {
        TopLevelPopCodeParameter.Value.Value = value;
      }
    }

    public IValueParameter<IntValue> MaxStringLengthParameter
    {
      get { return (IValueParameter<IntValue>)Parameters[MaxStringLengthParameterName]; }
    }

    public int MaxStringLength
    {
      get { return MaxStringLengthParameter.Value.Value; }
      set { MaxStringLengthParameter.Value.Value = value; }
    }
  }
}
